import threading
from typing import Callable, Any
import json
import time
from pynostr.event import Event
import requests
from pynostr.utils import get_public_key

from nostr_agents.nostr_client import NostrClient


class NostrAgentServer(object):
    def __init__(self, agent_url: str, satoshis: int, nostr_client: NostrClient):
        self.client = nostr_client
        self.agent_url = agent_url
        self.satoshis = satoshis
        self._agent_info = self._get_agent_info()

    def _get_agent_info(self) -> dict[str, Any]:
        return requests.get(f"{self.agent_url}/info",
                            headers={'Content-Type': 'application/json'},
                            ).json()

    def agent_info(self) -> dict[str, Any]:
        return self._agent_info

    def chat(
        self,
        message: str,
        thread_id: str | None = None,
    ) -> Any:
        """Call a tool by name with arguments."""
        request = {
            'messages': [message],
        }
        if thread_id:
            request['thread_id'] = thread_id
        print(f'Sending request: {json.dumps(request)}')
        response = requests.post(
            f"{self.agent_url}/chat",
            headers={'Content-Type': 'application/json'},
            json=request,
        )
        try:
            response.raise_for_status()
            result = response.text
        except Exception as e:
            print(f"Error: {e}")
            result = f'Unknown error'
        print(f'Response: {result}')
        return result

    def _direct_message_callback(self, event: Event, message: str):
        if event.pubkey != get_public_key('npub1jch03stp0x3fy6ykv5df2fnhtaq4xqvqlmpjdu68raaqcntca5tqahld7a').hex():
            print(f'Ignoring messages not from markcrypto')
            return
        """
        Callback function to handle incoming direct messages.
        :param event: The event object containing the message.
        :param message: The message content.
        """
        # Process the incoming message
        message = message.strip()
        print(f"Request: {message}")
        try:

            satoshis = self.satoshis
            if satoshis > 0:
                # Requires payment first
                invoice = self.client.nwc_client.make_invoice(amt=satoshis,
                                                              desc="Payment for agent")
                response = invoice

                def on_success():
                    print(f"Payment succeeded for agent")
                    result = self.chat(message, thread_id=event.pubkey)
                    response = {
                        "content": [{
                            "type": "text",
                            "text": str(result)
                        }]
                    }
                    print(f'On success response: {response}')
                    thr = threading.Thread(
                        target=self.client.send_direct_message_to_pubkey,
                        args=(event.pubkey, response),
                    )
                    thr.start()

                def on_failure():
                    response = {
                        "error": f"Payment failed"
                    }
                    print(f"On failure response: {response}")
                    thr = threading.Thread(
                        target=self.client.send_direct_message_to_pubkey,
                        args=(event.pubkey, json.dumps(response)),
                    )
                    thr.start()
                thr = threading.Thread(
                    target=self.client.nwc_client.on_payment_success,
                    kwargs={
                        'invoice': invoice,
                        'callback': on_success,
                        'timeout': 20,
                        'unsuccess_callback': on_failure,
                    }
                )
                thr.start()
            else:
                result = self.chat(message, thread_id=event.pubkey)
                response = {
                    "content": [{
                        "type": "text",
                        "text": str(result)
                    }]
                }
        except Exception as e:
            response = f'Error: {e}'

        print(f'Response: {response}')
        time.sleep(1)
        thr = threading.Thread(
            target=self.client.send_direct_message_to_pubkey,
            args=(event.pubkey, response),
        )
        thr.start()

    def start(self):
        # Update list_tools metadata
        thr = threading.Thread(
            target=self.client.update_metadata,
            kwargs={
                'name': 'agent_server',
                'display_name': self._agent_info['name'],
                'about': json.dumps(self.agent_info())
            }
        )
        print(f'Updating metadata for {self.client.public_key.bech32()}')
        thr.start()
        time.sleep(5)

        print(f'Starting message listener for {self.client.public_key.bech32()}')

        # Start call_tool listener
        self.client.direct_message_listener(
            callback=self._direct_message_callback
        )


if __name__ == "__main__":
    import os
    from dotenv import load_dotenv

    load_dotenv()

    # Get the environment variables
    relays = os.getenv('NOSTR_RELAYS').split(',')
    private_key = os.getenv('NOSTR_SERVER_PRIVATE_KEY')
    nwc_str = os.getenv('NWC_CONN_STR')
    agent_url = os.getenv('AGENT_URL')

    # Create an instance of NostrClient
    client = NostrClient(relays, private_key, nwc_str)
    server = NostrAgentServer(agent_url, 0, client)
    server.start()



from nostr_agents.nwc import processNWCstring, tryToPayInvoice, listTx, getBalance, didPaymentSucceed, makeInvoice, checkInvoice
from bolt11.decode import decode
import time


class NWCClient(object):
    def __init__(self, nwc_str: str):
        self.nwc_info = processNWCstring(nwc_str)

    def list_tx(self) -> list:
        return listTx(self.nwc_info)['result']['transactions']

    def get_balance(self) -> int:
        return getBalance(self.nwc_info)['result']['balance']

    def make_invoice(self, amt: int = None, desc: str = None) -> str:
        return makeInvoice(self.nwc_info, amt=amt, desc=desc)['result']['invoice']

    def check_invoice(self, invoice: str = None, payment_hash: str = None) -> dict:
        return checkInvoice(self.nwc_info, invoice=invoice, payment_hash=payment_hash)

    def did_payment_succeed(self, invoice: str = None) -> bool | str:
        return didPaymentSucceed(self.nwc_info, invoice=invoice)

    def try_pay_invoice(self, invoice: str, amt: int = None):
        decoded = decode(invoice)
        if decoded.amount_msat and amt:
            if decoded.amount_msat != amt * 1000:  # convert to msats
                raise RuntimeError(f'Amount in invoice [{decoded.amount_msat}] does not match amount provided [{amt}]')
        elif not decoded.amount_msat and not amt:
            raise RuntimeError('No amount provided in invoice and no amount provided to pay')
        tryToPayInvoice(self.nwc_info, invoice=invoice, amnt=amt)

    def on_payment_success(self, invoice: str, callback=None, unsuccess_callback=None, timeout: int = 60, interval: int = 5):
        """
        Listen for payment success for a given invoice.
        :param invoice: The invoice to listen for.
        :param callback: The callback to call on success.
        :param timeout: The timeout in seconds.
        :param interval: The interval in seconds to check for payment success.
        """
        start_time = time.time()
        success = False
        while True:
            if self.did_payment_succeed(invoice):
                success = True
                if callback:
                    callback()
                break
            if time.time() - start_time > timeout:
                break
            time.sleep(interval)
        if not success:
            if unsuccess_callback:
                unsuccess_callback()



import threading
from typing import Any
import json

from pynostr.event import Event
from pynostr.key import PrivateKey

from nostr_agents.nostr_client import NostrClient


class NostrMCPClient(object):
    def __init__(self, nostr_client: NostrClient, mcp_pubkey: str):
        self.mcp_pubkey = mcp_pubkey
        self.client = nostr_client
        self.tool_to_sats_map = {}

    def _set_result_callback(self, tool_name: str, res: list):
        def inner(event: Event, message: str):
            try:
                res[0] = json.loads(message)
                if invoice := res[0].get('invoice'):
                    print(f'Paying invoice: {invoice}')
                    self.client.nwc_client.try_pay_invoice(invoice=invoice, amt=self.tool_to_sats_map[tool_name])
                    return False  # Keep listening
                return True
            except Exception as e:
                print(f"Error parsing message: {e}")
            return False
        return inner

    def list_tools(self) -> dict[str, Any] | None:
        """Retrieve available tools"""
        metadata = self.client.get_metadata_for_pubkey(self.mcp_pubkey)
        tools = json.loads(metadata.about)
        for tool in tools['tools']:
            self.tool_to_sats_map[tool['name']] = tool['satoshis']
        return tools

    def call_tool(
        self,
        name: str,
        arguments: dict[str, Any],
    ) -> dict[str, Any] | None:
        """Call a tool by name with arguments (paying satoshis if required)."""
        thr = threading.Thread(
            target=self.client.send_direct_message_to_pubkey,
            args=(self.mcp_pubkey, json.dumps({
                'action': 'call_tool',
                'tool_name': name,
                'arguments': arguments
            })),
        )
        thr.start()
        res = [None]
        self.client.direct_message_listener(
            callback=self._set_result_callback(name, res),
            recipient_pubkey=self.mcp_pubkey,
            timeout=30,
            close_after_first_message=True
        )
        return res[0]


if __name__ == "__main__":
    import os
    from dotenv import load_dotenv

    load_dotenv()

    # Get the environment variables
    relays = os.getenv('NOSTR_RELAYS').split(',')
    private_key = os.getenv('NOSTR_CLIENT_PRIVATE_KEY')
    server_public_key = PrivateKey.from_nsec(os.getenv('NOSTR_SERVER_PRIVATE_KEY')).public_key.hex()
    nwc_str = os.getenv('NWC_CONN_STR')

    print(f"Server public key: {PrivateKey.from_nsec(os.getenv('NOSTR_SERVER_PRIVATE_KEY')).public_key.bech32()}")

    # Create an instance of NostrClient
    client = NostrClient(relays, private_key, nwc_str)
    mcp_client = NostrMCPClient(client, mcp_pubkey=server_public_key)

    tools = mcp_client.list_tools()
    print(f'Found tools:')
    print(json.dumps(tools, indent=4))

    result = mcp_client.call_tool("get_weather", {"city": "Seattle"})
    print(f'Result: {result}')

    print(f'The weather in Seattle is: {result["content"][-1]["text"]}')

    result = mcp_client.call_tool("multiply", {"a": 69, "b": 420})
    print(f'The result of 69 * 420 is: {result["content"][-1]["text"]}')

    result = mcp_client.call_tool("get_current_date", {})
    print(f'The current date is: {result["content"][-1]["text"]}')
